import hashlib
from pathlib import Path

from tqdm import tqdm

from . import constants


class GeneratedSet:
    """
    Bookkeeping for generation of thumbnails and contact sheets.
    """

    def __init__(self, generated_dir):
        self.source_images = []
        self.generated_dir = generated_dir
        # keys are Paths, values are digests
        self.digest_cache = {}

    def register_source(self, image):
        """registers an image for thumnail generation."""
        self.source_images.append(image)

    def divergence(self) -> tuple[list[Path], list[Path]]:
        obsolete = set(self.generated_dir.glob(self.generated_file_prefix() + '-*'))
        missing = []
        for f in self.source_images:
            generated_path = self.generated_path(f)
            obsolete.discard(generated_path)
            if not generated_path.exists():
                missing.append(f)
        return missing, list(obsolete)

    def digest(self, path):
        if path not in self.digest_cache:
            with open(path, 'rb', buffering=0) as f:
                digest = hashlib.file_digest(f, constants.digest).hexdigest()
            self.digest_cache[path] = digest
        return self.digest_cache[path]

    def generated_path(self, fpath):
        digest = self.digest(fpath)
        basename = "%s-%s%s" % (
            self.generated_file_prefix(),
            digest,
            self.generated_file_suffix(fpath))
        return self.generated_dir / basename

    def generated_file_suffix(self, fpath):
        """Default to the same suffix as the source file"""
        return fpath.suffix

    def generated_file_prefix(self):
        raise NotImplementedError

    def create_missing(self, missing):
        self.generated_dir.mkdir(exist_ok=True, parents=True)
        for source in tqdm(missing):
            destination = self.generated_dir / self.generated_path(source)
            # print("creating", source, destination)
            self.create_file(source, destination)

    def unlink_obsolete(self, obsolete):
        for path in obsolete:
            path.unlink()

    def create_file(self, source, destination):
        raise NotImplementedError



import argparse
from pathlib import Path

from . import constants, data_types, dynimport, gallery

from .renderer import Renderer
gallery_name_help = '''
The name of the gallery. Defaults the base name of the 'source_dir'.
'''

source_dir_help = f'''
The directory containing the source images and videos over which we want to
create a gallery.
'''

gallery_dir_help = '''
The directory in which to store the generated gallery. Defaults to the same as
the 'source_dir' containing the images.

Note that this directory should be "close" to the `source_dir` since relative
paths are used when referencing source images from the gallery or you'll get
many '../' elements in the image paths.
'''

recursive_help = '''
Whether to search for image and video files recursively.
'''

renderer_help = '''
Which renderer to use to actually produce the output galleries. At the moment,
there are two built-in ones: "PhotoSwipe" and "nanogallery2". Advanced: Other
values will be loaded as a module that is expected to have a renderer() method
that returns an instance of gallerator.data_types.Renderer. That way you can
render the gallery exactly like you want.
'''

pagination_help = '''
The maximum number of images per page. 0 means unlimited.
'''


def get_renderer(renderer_arg):
    def get_renderer(path):
        if not Path(path).exists():
            raise FileNotFoundError(path)
        module = dynimport.import_random_module_from_path(path)
        if not hasattr(module, 'renderer'):
            raise ValueError(
                f'Renderer {path} does not have a renderer method')
        renderer = module.renderer()
        if not isinstance(renderer, Renderer):
            raise ValueError(
                f'Renderer {path}.render() does not return a Renderer instance')
        return renderer

    match renderer_arg:
        case "PhotoSwipe":
            return get_renderer(
                Path(__file__).parent / "renderers" /
                "photoswipe" / "photoswipe.py")
        case "nanogallery2":
            return get_renderer(
                Path(__file__).parent / "renderers" /
                "nanogallery2" / "nanogallery2.py")
        case _:
            return get_renderer(renderer_arg)


def parse_args():
    def add_renderer_argument(parser):
        parser.add_argument(
            '--renderer', default='PhotoSwipe', help=renderer_help)

    preparser = argparse.ArgumentParser(
        prog='gallerator',
        add_help=False
    )
    add_renderer_argument(preparser)
    preargs, _ = preparser.parse_known_args()
    renderer = get_renderer(preargs.renderer)

    parser = argparse.ArgumentParser(
        prog='gallerator',
        description='Create static thumbnail galleries',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('source_dir', help=source_dir_help)
    parser.add_argument('--name-of-gallery',
                        help=gallery_name_help, default=None)
    parser.add_argument('--gallery-dir', '-g',
                        help=gallery_dir_help, default=None)
    parser.add_argument('--recursive', '-r', help=recursive_help,
                        default=False, action='store_true')
    parser.add_argument('--pagination', '-p',
                        type=int,
                        default=0, help=pagination_help)
    add_renderer_argument(parser)

    renderer.add_argparse_args(parser)

    args = parser.parse_args()

    renderer.update_args(args)

    return args, renderer


def cli_main():
    args, renderer = parse_args()
    if args.name_of_gallery is None:
        gallery_name = Path(args.source_dir).stem
    else:
        gallery_name = args.name_of_gallery
    if args.gallery_dir is None:
        gallery_dir = args.source_dir
    else:
        gallery_dir = args.gallery_dir
    gallery.write_gallery(
        gallery_name,
        Path(args.source_dir).absolute(),
        Path(gallery_dir).absolute(),
        args.recursive,
        renderer,
        args.pagination)



import hashlib
import os
import tempfile
import unittest
from pathlib import Path

from gallerator import constants, generated_set


def expected_file_digest(path):
    with open(path, 'rb', buffering=0) as f:
        digest = hashlib.file_digest(f, constants.digest).hexdigest()
    return digest


class TestGeneratedSetImpl(generated_set.GeneratedSet):
    """Test instance of generated set that implements generated_path()"""

    def generated_file_prefix(self):
        return "test"

    def create_file(self, source, destination):
        with open(destination, 'w') as f:
            f.write(f'A file for {source}')


class GeneratedSetTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.source_path = Path(__file__).parent / '../../../../demo-gallerator/media'
        self.sample_images = list(self.source_path.glob('*.jpg'))
        self.sample_videos = sorted(
            list((self.source_path / 'Videos').glob('*.mp4')))
        self.assertTrue(self.source_path.exists())
        self.assertGreater(len(self.sample_images), 0)
        self.assertGreater(len(self.sample_videos), 0)

        self.tempdir = tempfile.TemporaryDirectory(
            prefix="gallerator-" + self._testMethodName,
            delete=not (self.keep_test_data()))
        self.output_dir = Path(self.tempdir.name)
        if self.keep_test_data():
            print('Keeping test data in %s' % self.output_dir)

        return super().setUp()

    def tearDown(self) -> None:
        if not (self.keep_test_data()):
            self.tempdir.cleanup()
        return super().tearDown()

    def keep_test_data(self):
        return "KEEPTESTDATA" in os.environ and os.environ["KEEPTESTDATA"]


class TestGeneratedSet(GeneratedSetTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.generatedset = TestGeneratedSetImpl(self.output_dir)

    def test_digests(self):
        expected = [expected_file_digest(
            self.source_path / f) for f in self.sample_images[0:2]]
        actual = [self.generatedset.digest(
            f) for f in self.sample_images[0:2]]
        self.assertEqual(actual, expected)

    def test_divergence(self):
        bogus_test_file = self.output_dir / \
            f'test-{expected_file_digest(__file__)}.txt'
        with open(bogus_test_file, 'w') as f:
            f.write('bogus contents')
        for i in range(2):
            self.generatedset.register_source(self.sample_images[i])
        missing, obsolete = self.generatedset.divergence()
        self.assertEqual(
            (missing, obsolete,),
            ([self.output_dir / f for f in self.sample_images[0:2]],
             [bogus_test_file],)
        )
        self.generatedset.create_missing(missing)
        self.generatedset.unlink_obsolete(obsolete)
        missing, obsolete = self.generatedset.divergence()
        self.assertEqual(
            (missing, obsolete,),
            ([], [],)
        )

    def test_paths(self):
        expected = []
        for f in self.sample_images[0:2]:
            digest = expected_file_digest(self.source_path / f)
            path = "%s/test-%s.jpg" % (self.output_dir, digest)
            expected.append(Path(path))
            # self.generatedset.thumbnail_path
        actual = [self.generatedset.generated_path(
            f) for f in self.sample_images[0:2]]
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Callable


class MediaType(Enum):
    UNKNOWN = 1
    IMAGE = 2
    VIDEO = 3


@dataclass
class ImageInfo:
    path: Path
    width: int
    height: int

@dataclass
class MediaItem:
    """
        Representation of a gallery media item. Uses absolute paths for files.
    """
    type: MediaType
    title: str
    thumbnail: ImageInfo
    image: ImageInfo
    source: Path
    video: Path | None


@dataclass
class Directory:
    """
        Representation of a gallery directory. Uses absolute paths for files.
    """
    name: str
    path_segments: list[str]
    items: list[MediaItem] = field(default_factory=list)
    subdirectories: "dict[Directory]" = field(default_factory=dict)

@dataclass
class Breadcrumb:
    name: str
    path_segments: list[str]

@dataclass
class TemplateVars:
    gallery_name: str
    gallery_root_url: str
    parent: Directory | None
    # Each element contains a name and a url
    breadcrumbs: list[Breadcrumb]
    media_items: list[MediaItem]
    subdirectories: list[Directory]
    path_segments_to_url: Callable[[list[Path]],str]
    relative_url: Callable[[Path],str]
    # 0 based - first page has page_num = 0
    page_num: int
    # total_pages == 1 (and page_num = 0) if pagination is disabled
    total_pages: int
    url_for_page_num: Callable[[int],str]



from pathlib import Path

from . import constants, data_types, media_items, url_strategy


def create_template_vars_plural(
    gallery_name: str,
    gallery_path: Path,
    directory: data_types.Directory,
    parent: data_types.Directory,
    page_url_strategy: url_strategy.PageUrlStrategy,
    pagination: int,
):
    subdirectories = []
    for dir in sorted(directory.subdirectories):
        subdirectories.append(directory.subdirectories[dir])

    breadcrumbs = [
        data_types.Breadcrumb(
            name=gallery_name,
            path_segments=[]
        )
    ]
    path_segments_so_far = []
    for p in directory.path_segments:
        path_segments_so_far.append(p)
        breadcrumbs.append(data_types.Breadcrumb(
            name=p,
            path_segments=path_segments_so_far.copy(),
        ))

    # This is the absolute linux path of the current "directory", not the
    # current directory of the running python process.
    current_dir = (
        gallery_path / page_url_strategy.page_url(directory.path_segments)
    ).parent

    def path_segments_to_url(path_segments: list[Path]) -> str:
        abs_page = gallery_path / page_url_strategy.page_url(path_segments)
        return str(abs_page.relative_to(current_dir, walk_up=True))

    def relative_url(path: Path) -> str:
        return str(path.relative_to(current_dir, walk_up=True))

    def url_for_page_num(page_num: int) -> str:
        abs_page = gallery_path / page_url_strategy.page_url(
            directory.path_segments, page_num
        )
        return str(abs_page.relative_to(current_dir, walk_up=True))

    current_media_item_index = 0
    template_vars_plural = []

    def append_template_vars_plural(items, page_num, total_pages):
        template_vars = data_types.TemplateVars(
            gallery_name=gallery_name,
            gallery_root_url=str(gallery_path.relative_to(
                current_dir, walk_up=True)),
            parent=parent,
            breadcrumbs=breadcrumbs,
            media_items=items,
            subdirectories=subdirectories,
            path_segments_to_url=path_segments_to_url,
            relative_url=relative_url,
            page_num=page_num,
            total_pages=total_pages,
            url_for_page_num=url_for_page_num,
        )
        template_vars_plural.append(template_vars)

    # Some very large number to ensure we don't create more than one page
    if pagination is None or pagination == 0:
        pagination = 100000000

    if len(directory.items) == 0:
        # Every directory needs at least one page
        append_template_vars_plural([], 0, 1)
    else:
        total_pages = 1 + int(len(directory.items) / pagination)
        page_num = 0
        while current_media_item_index < len(directory.items):
            remaining = len(directory.items) - current_media_item_index
            nr_items = min(remaining, pagination)
            items = directory.items[current_media_item_index:
                                    current_media_item_index + nr_items]
            append_template_vars_plural(items, page_num, total_pages)
            current_media_item_index += nr_items
            page_num += 1

    return template_vars_plural


def write_gallery_directory(renderer,
                            gallery_name: str,
                            gallery_path: Path,
                            directory: data_types.Directory,
                            parent: data_types.Directory | None,
                            page_url_strategy: url_strategy.PageUrlStrategy,
                            pagination: int | None):
    template_vars_plural = create_template_vars_plural(
        gallery_name,
        gallery_path,
        directory,
        parent,
        page_url_strategy,
        pagination,
    )
    for template_vars in template_vars_plural:
        fname = gallery_path / \
            page_url_strategy.page_url(
                directory.path_segments, template_vars.page_num)
        fname.parent.mkdir(exist_ok=True)
        print(f'Creating {fname}')
        html = renderer.render(template_vars)
        with open(fname, 'w') as file:
            file.write(html)
    # recursive!
    for subdir in directory.subdirectories:
        write_gallery_directory(renderer,
                                gallery_name,
                                gallery_path,
                                directory.subdirectories[subdir],
                                directory,
                                page_url_strategy,
                                pagination)


def write_gallery(
        gallery_name,
        src_path,
        gallery_path,
        recursive,
        renderer,
        pagination):

    root = media_items.create_directory_media(
        src_path, gallery_path, recursive)

    # page_url_strategy = url_strategy.UnderscorePageUrlStrategy()
    page_url_strategy = url_strategy.DirectoryPageUrlStrategy()

    write_gallery_directory(
        renderer,
        gallery_name,
        gallery_path,
        root,
        None,
        page_url_strategy,
        pagination
    )

    print(f"Copying static files to {gallery_path}")
    renderer.copy_static(gallery_path)



import math
import sys
from contextlib import redirect_stdout

from vcsi.vcsi import MediaInfo
from vcsi.vcsi import main as vcsi_main

from . import constants, generated_set


def run_vcsi(*args):
    old_sys_argv = sys.argv
    sys.argv = ['vcsi'] + [str(s) for s in args]
    try:
        # vcsi is noisy and prints progress to stdout. Don't do that.
        with redirect_stdout(None):
            vcsi_main()
    finally:
        sys.argv = old_sys_argv


class VideoSamples(generated_set.GeneratedSet):

    def __init__(self, generated_dir):
        super().__init__(generated_dir)

    def generated_file_prefix(self):
        return 'video-sample'

    def generated_file_suffix(self, fpath):
        return ".jpg"

    def create_file(self, source, destination):
        # If we just let vcsi create a single image, it will do so towards the
        # middle of the video. Testing shows that the first image in the contact
        # sheet, about 12% into the video, is better. Find the time 12% in. For
        # that we could use ffprobe (like vcsi does), but it already does
        # everything so well, so why not let it...
        #
        # That being said, we calculate the width that should make the height
        # constants.thumbnail_height (300) but the result is actually 395
        # (landscape video) or 413 (portait) so I guess it is good enough. But
        # not precise...
        media_info = MediaInfo(source)
        thumbnail_seconds = media_info.duration_seconds * \
            constants.video_thumbnail_time_fraction
        minutes = int(thumbnail_seconds / 60)
        seconds = int(thumbnail_seconds) % 60
        ratio = math.sqrt(
            media_info.display_height * media_info.display_width
            / constants.thumbnail_target_pixels)
        width = media_info.display_width / ratio
        run_vcsi(source,
                 '-t',
                 '-g', '1x1',
                 '--width', int(width),
                 '--metadata-position', 'hidden',
                 # We can't seem to hide timestamps completely, but this
                 # makes them "very small".
                 '--timestamp-format', '',
                 # It doesn't look to me like --timestamp-font-size works at
                 # all
                 '--timestamp-font-size', '40',
                 '-m', f'{minutes}:{seconds}',
                 '-o', destination)


class VideoContactSheets(generated_set.GeneratedSet):

    def __init__(self, generated_dir):
        super().__init__(generated_dir)

    def generated_file_prefix(self):
        return 'video-contact-sheet'

    def generated_file_suffix(self, fpath):
        return ".jpg"

    def create_file(self, source, destination):
        run_vcsi(source, '-t', '-o', destination)



from gallerator import video_images
from gallerator.tests.test_generated_set import GeneratedSetTestCase


class TestVideoSamples(GeneratedSetTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.video_samples = video_images.VideoSamples(self.output_dir)

    def test_create_missing_video_samples(self):
        for i in range(2):
            self.video_samples.register_source(self.sample_videos[i])
        missing, obsolete = self.video_samples.divergence()
        self.video_samples.create_missing(missing)
        for i in range(2):
            path = self.video_samples.generated_path(self.sample_videos[i])
            self.assertTrue(path.exists())
        self.assertEqual(
            self.video_samples.divergence(),
            ([], [],)
        )

class TestVideoContactSheets(GeneratedSetTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.video_contact_sheets = \
            video_images.VideoContactSheets(self.output_dir)

    def test_create_missing_contact_sheets(self):
        for i in range(2):
            self.video_contact_sheets.register_source(self.sample_videos[i])
        missing, obsolete = self.video_contact_sheets.divergence()
        self.video_contact_sheets.create_missing(missing)
        for i in range(2):
            path = self.video_contact_sheets.generated_path(self.sample_videos[i])
            self.assertTrue(path.exists())
        self.assertEqual(
            self.video_contact_sheets.divergence(),
            ([], [],)
        )



import re
from pathlib import Path


class PageUrlStrategy:
    def page_url(self, paths: list[Path], page_num: int = 0) -> str:
        """page_num is zero based"""
        raise NotImplementedError


class DirectoryPageUrlStrategy(PageUrlStrategy):
    def page_url(self, paths: list[Path], page_num: int = 0) -> str:
        """page_num is zero based"""
        local_paths = paths.copy()
        if page_num == 0:
            local_paths.append('index.html')
        else:
            local_paths.append(f'page{page_num+1:03}.html')
        return "/".join(list(local_paths))


class UnderscorePageUrlStrategy(PageUrlStrategy):
    def page_url(self, paths: list[Path], page_num: int = 0) -> str:
        """page_num is zero based"""

        if len(paths) == 0:
            prefix = 'index'
        else:
            # Escape any ":" chars in directory names
            prefix = ("_".join([re.sub(r"_", "__", p) for p in paths]))
        if page_num > 0:
            prefix += f'_page{page_num+1:03}'
        return prefix + '.html'



import unittest

from gallerator import url_strategy


class DirectoryPageUrlStrategyTest(unittest.TestCase):
    def setUp(self):
        self.strategy = url_strategy.DirectoryPageUrlStrategy()

    def test_root(self):
        self.assertEqual(
            self.strategy.page_url([]),
            'index.html'
        )

    def test_page_num_root(self):
        self.assertEqual(
            self.strategy.page_url([], 2),
            'page003.html'
        )
        
    def test_single_path(self):
        self.assertEqual(
            self.strategy.page_url(['singlepath']),
            'singlepath/index.html'
        )

    def test_single_path_underscore(self):
        self.assertEqual(
            self.strategy.page_url(['single_path']),
            'single_path/index.html'
        )

    def test_multiple_path_underscore(self):
        self.assertEqual(
            self.strategy.page_url(['path_1', 'path_2']),
            'path_1/path_2/index.html'
        )

    def test_page_num(self):
        self.assertEqual(
            self.strategy.page_url(['path_1', 'path_2'], 2),
            'path_1/path_2/page003.html'
        )

class UnderscorePageUrlStrategyTest(unittest.TestCase):
    def setUp(self):
        self.strategy = url_strategy.UnderscorePageUrlStrategy()

    def test_root(self):
        self.assertEqual(
            self.strategy.page_url([]),
            'index.html'
        )

    def test_page_num_root(self):
        self.assertEqual(
            self.strategy.page_url([], 2),
            'index_page003.html'
        )
        
    def test_single_path(self):
        self.assertEqual(
            self.strategy.page_url(['singlepath']),
            'singlepath.html'
        )

    def test_single_path_underscore(self):
        self.assertEqual(
            self.strategy.page_url(['single_path']),
            'single__path.html'
        )

    def test_multiple_path_underscore(self):
        self.assertEqual(
            self.strategy.page_url(['path_1', 'path_2']),
            'path__1_path__2.html'
        )

    def test_page_num(self):
        self.assertEqual(
            self.strategy.page_url(['path_1', 'path_2'], 2),
            'path__1_path__2_page003.html'
        )



from pathlib import Path

import argparse
import re

from gallerator import constants, data_types, renderer
from gallerator.renderers import renderer_util

description = """
These are the arguments for Photoswipe, the default renderer. Different
renderers have different options, so try "--renderer name-of-renderer --help" to
see options for other renderers.
"""

justified_help = """
Use a justified layout (the default), where all images have the same height, but
the width depends on the image. This is the most dense but different pages may
not have the same height if used with pagination, so perhaps one other layouts
are better with pagination.
"""

auto_grid_help = """
Use a grid layout but fit as many columns as will fit depending on screen width.
"""

grid_help = """
By default, Photoswipe uses a justified layout, but this will create a grid
layout. Try e.g. "--grid 4x5", for 4 columns by 5 rows, but supply your own
"XxY" value to set your own grid layout. This will override any --pagination
value. Beware that this is not responsive, so it will either be too wide on
mobile or too narrow for large desktop users, or both.
"""

width_help = """
The width of the thumbnails. Only used by --auto-grid and --grid
"""

height_help = """
The height of the thumbnails.
"""
no_filename_captions_help = """
Don't add the filenames of the images as a caption in the thumbnail gallery.
This makes it look cleaner (especially for --grid and --auto-grid), but less
information for the user. When the full image is shown in a lightbox, the file
name is always shown, regardless of this setting.
"""


class Photoswipe(renderer.Renderer):
    def __init__(self):
        self.jenv = renderer_util.jinja2_env(
            Path(__file__).parent / 'templates')

    def add_argparse_args(self, parser: argparse.ArgumentParser):
        group = parser.add_argument_group(
            'Photoswipe', description=description)
        group.add_argument('--justified', action='store_true', help=grid_help)
        group.add_argument('--auto-grid', action='store_true', help=grid_help)
        group.add_argument('--grid', help=grid_help)
        group.add_argument('--width', default=300, help=width_help)
        group.add_argument('--height', default=300, help=height_help)
        group.add_argument('--no-filename-captions',
                           action="store_true", help=no_filename_captions_help)

    def _set_grid_dimensions(self, args):
        self.grid = None
        if args.grid is not None:
            matches = re.match(r'^(\d+)x(\d+)$', args.grid)
            if matches is None:
                raise ValueError(
                    "Please sepecify --grid as XxY, where X and Y are "
                    "integers, e.g. 4x5"
                )
            x = int(matches[1])
            y = int(matches[2])
            args.pagination = x * y
            self.grid = {
                "type": "fixed-columns",
                "x": x,
                "y": y,
            }
        elif args.auto_grid:
            self.grid = {
                "type": "auto",
            }

    def update_args(self, args: argparse.Namespace):
        self._set_grid_dimensions(args)
        super().update_args(args)

    def render(self, template_vars: data_types.TemplateVars):
        pagination = renderer_util.pagination_controls(
            template_vars.page_num,
            template_vars.total_pages,
            template_vars.url_for_page_num
        )
        template = self.jenv.get_template("page.html")
        vars = template_vars.__dict__.copy()
        vars.update({
            'data_types': data_types,
            'layout': 'justified' if self.grid is None else 'grid',
            'pagination': pagination,
        })
        return template.render(vars)

    def copy_static(self, gallery_path: Path):
        renderer_util.copy_static(
            Path(__file__).parent / 'static', gallery_path)
        template = self.jenv.get_template('page.css')
        with open(gallery_path / 'static' / 'page.css', 'w') as f:
            f.write(template.render(
                thumbnail_height=self.args.height,
                thumbnail_width=self.args.width,
                grid=self.grid,
                filename_captions=not(self.args.no_filename_captions)
            ))


def renderer() -> renderer.Renderer:
    return Photoswipe()



generated_dir_basename = 'generated-images'

digest = 'sha1'

thumbnail_target_pixels = 350*350

# If we just let vcsi create a single image, it will do so towards the
# middle of the video. Testing shows that the first image in the contact
# sheet, about 12% into the video, is better. Find the time 12% in:
video_thumbnail_time_fraction = 0.12


from dataclasses import dataclass
from pathlib import Path
from typing import Callable

from PIL import Image

from . import constants, data_types, video_images
from . import thumbnails as thumbnails_module

image_suffixes = {'.' + ext for ext in {
    "png",
    "jpg",
    "jpeg",
    "bmp",
    "gif",
    "svg"
}}

video_suffixes = {'.' + ext for ext in {
    "avi",
    "mp4",
    "wmv",
    "mov",
    "ram"
}}


def determine_media_type(path):
    """Determine the media type of path, perhaps being UKNOWN"""
    suffix = path.suffix.lower()
    if suffix in image_suffixes:
        return data_types.MediaType.IMAGE
    elif suffix in video_suffixes:
        return data_types.MediaType.VIDEO
    else:
        return data_types.MediaType.UNKNOWN

def determine_known_media_type(path):
    """Determine the media type of path, UKNOWN being an error"""
    mediatype = determine_media_type(path)
    if mediatype == data_types.MediaType.UNKNOWN:
        raise ValueError(f"{path} does not have a known file extension")
    return mediatype



def find_media(src_path: Path, recursive: bool) -> list[Path]:
    media = []
    for root, dirs, files in src_path.walk():
        if recursive:
            # Skip over any _generated-images directories so we don't make
            # thumbnails of thumnails ad nauseam.
            if constants.generated_dir_basename in dirs:
                dirs.remove(constants.generated_dir_basename)
            # Skip over any static directories containing a favicon.ico,
            # assuming they've been created by the gallery on a previous run.
            # Yeah, this is a little ugly, with a hardcoded file name and all. I
            # just couldn't bring myself to make 'static' a constant and
            # incorporate it in templates and everything. So it is what it is.
            # Patches welcome.
            if 'static' in dirs and (root / 'static' / 'favicon.ico').exists():
                dirs.remove('static')
        else:
            # If not recursive, don't go into any subdirs
            dirs.clear()
        for f in files:
            fpath = root / f
            if determine_media_type(fpath) == data_types.MediaType.UNKNOWN:
                continue
            media.append(fpath)
    return sorted(media)


def group_media_in_directories(media: list[Path], src_path):
    directories = {}
    for path in sorted(media):
        dir = str(path.parent.relative_to(src_path))
        # special case for the root of the tree
        if dir == '.':
            dir = ''
        directories[dir] = directories.get(dir, []) + [path]
    return directories


@dataclass
class DerivedMedia:
    thumbnails: thumbnails_module.Thumbnails
    video_samples: video_images.VideoSamples
    video_contact_sheets: video_images.VideoContactSheets


def register_derived_media(media, generated_dir):
    thumbnails = thumbnails_module.Thumbnails(generated_dir)
    video_samples = video_images.VideoSamples(generated_dir)
    video_contact_sheets = video_images.VideoContactSheets(generated_dir)
    for path in media:
        match determine_known_media_type(path):
            case data_types.MediaType.IMAGE:
                thumbnails.register_source(path)
            case data_types.MediaType.VIDEO:
                video_samples.register_source(path)
                video_contact_sheets.register_source(path)
            case _:
                raise RuntimeError(f"Unknown type for {path}")
    return DerivedMedia(
        thumbnails=thumbnails,
        video_samples=video_samples,
        video_contact_sheets=video_contact_sheets)


def create_image_info(abs_path):
    im = Image.open(abs_path)
    width, height = im.size
    return data_types.ImageInfo(
        path = abs_path,
        width = width,
        height = height,
    )



def create_media_items(media, derived_media, gallery_path):
    items = []
    for path in media:
        type = determine_known_media_type(path)
        match type:
            case data_types.MediaType.IMAGE:
                thumbnail = derived_media.thumbnails.generated_path(path)
                image = path
                video = None
            case data_types.MediaType.VIDEO:
                thumbnail = derived_media.video_samples.generated_path(path)
                image = derived_media.video_contact_sheets.generated_path(path)
                video = path
            case _:
                raise ValueError(f"Uknown type for {path}")
        item = data_types.MediaItem(
            type=type,
            title=path.name,
            thumbnail=create_image_info(thumbnail),
            image=create_image_info(image),
            source=path,
            video=video
        )
        items.append(item)
    return items


def create_missing_media(derived_media):
    for (disp, obj) in [
        ('Thumbnails', derived_media.thumbnails),
        ('Video Samples', derived_media.video_samples),
        ('Video Contact Sheets', derived_media.video_contact_sheets),
    ]:
        missing, obsolete = obj.divergence()
        if len(missing) > 0:
            print(f"Creating {len(missing)} missing {disp}")
            obj.create_missing(missing)
        if len(obsolete) > 0:
            print(f"Removing {len(obsolete)} obsolete {disp}")
            obj.unlink_obsolete(obsolete)


def create_directory_media(
    src_path: Path,
    gallery_path: Path,
    recursive=True):
    media = find_media(src_path, recursive)
    generated_dir = gallery_path / constants.generated_dir_basename
    derived_media = register_derived_media(media, generated_dir)
    create_missing_media(derived_media)
    directory_media = group_media_in_directories(media, src_path)
    flat_directory_items = {}
    for directory in directory_media:
        flat_directory_items[directory] = create_media_items(
            directory_media[directory], derived_media, gallery_path)

    directory_names = sorted(list(directory_media.keys()))
    if '' in flat_directory_items:
        root_items = flat_directory_items['']
    else:
        root_items = []
    root = data_types.Directory(name='', path_segments=[], items=root_items)
    for dir in directory_names:
        # already setup root
        if dir == '':
            continue
        current_dir = root
        current_path_segments = []
        for part in Path(dir).parts:
            current_path_segments.append(part)
            if part not in current_dir.subdirectories:
                current_dir.subdirectories[part] = data_types.Directory(
                    name=part, path_segments=current_path_segments.copy()
                )
            current_dir = current_dir.subdirectories[part]
        current_dir.items = flat_directory_items[dir]
    return root


if __name__ == "__main__":
    sample = Path(__file__).parent.parent.parent / "sample"
    gallery_path = sample.parent / 'gallery'
    root = create_directory_media(sample, gallery_path, True)

    print(root)



import importlib.util
import random
import string
import sys


# https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
def import_from_path(module_name, file_path):
    if not file_path.exists():
        raise FileNotFoundError(file_path)
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module

# I don't care about the module name, so create a random one.
def import_random_module_from_path(file_path):
    module = 'dynimport_' + ''.join(random.choices(string.ascii_uppercase, k=20))
    return import_from_path(module, file_path)

if __name__ == "__main__":
    file = sys.argv[1]
    mod = import_random_module_from_path(file)
    print(mod.render())




import unittest

from gallerator import thumbnails
from gallerator.tests.test_generated_set import GeneratedSetTestCase


class TestThumbnails(GeneratedSetTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.thumbnails = thumbnails.Thumbnails(self.output_dir)

    def test_create_missing(self):
        for i in range(2):
            self.thumbnails.register_source(self.sample_images[i])
        missing, obsolete = self.thumbnails.divergence()
        self.thumbnails.create_missing(missing)
        def expected1(f):
            digest = self.thumbnails.digest(f)
            return self.output_dir / ('thumb-%s.jpg' % digest)
        expected = [expected1(f) for f in self.sample_images[0:2]]
        actual = list(self.output_dir.glob('*'))
        self.assertEqual(actual, expected)
        self.assertEqual(
            self.thumbnails.divergence(),
            ([], [],)
        )

if __name__ == '__main__':
    unittest.main()



import argparse
import json
import sys
from pathlib import Path

from gallerator import data_types, renderer
from gallerator.renderers import renderer_util

description = """
These are the arguments for the nanogallery2 renderer.
Beware of its license.
"""


height_help = """
The height of the thumbnails.
"""

not_downloaded_message = """
**************************************************************
* !!! Nanogallery2 distribution files could not be found !!! *
**************************************************************

It looks like the nanogallery2 distribution files have not
been downoaded. They are needed but not included with this
distribution due to their GPLv3 license.

If you have the source checked out, you can run

    make download-nanogallery2

from the top-level directory, or run

    $module)location/renderers/nanogallery2/download.sh

if you have installed the module with by some other means (e.g. pip), this could
be something like:

    ./venv/lib/python3.12/site-packages/gallerator/renderers/nanogallery2/download.sh
"""

def json_dumps_media_items(media_items, relative_url):
    items = []
    for i in media_items:
        item = {
            "title": i.title,
            "src": relative_url(i.image.path),
            "srct": relative_url(i.thumbnail.path),
        }
        if i.video is not None:
            item['customData'] = {"video": relative_url(i.video)}
        items.append(item)
    return json.dumps(items, indent=4)


class Nanongallery2(renderer.Renderer):
    def __init__(self):
        self.jenv = renderer_util.jinja2_env(
            Path(__file__).parent / 'templates')

    def add_argparse_args(self, parser: argparse.ArgumentParser):
        group = parser.add_argument_group(
            'nanogallery2', description=description)
        group.add_argument('--height', default=300, help=height_help)

    def render(self, template_vars: data_types.TemplateVars):
        template = self.jenv.get_template("page.html")
        render_vars = template_vars.__dict__
        render_vars["media_items_json"] = json_dumps_media_items(
            template_vars.media_items,
            template_vars.relative_url,
        )
        render_vars['thumbnail_height'] = self.args.height
        return template.render(render_vars)

    def copy_static(self, gallery_path: Path):
        static = Path(__file__).parent / 'static'
        if not (static / 'jquery.nanogallery2.min.js').exists():
            print(not_downloaded_message, file=sys.stderr)            
            sys.exit(1)
        renderer_util.copy_static(
            static, gallery_path)


def renderer() -> renderer.Renderer:
    return Nanongallery2()



import shutil
from pathlib import Path

from jinja2 import (
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    select_autoescape,
)

from gallerator import data_types


def jinja2_env(renderer_templates):
    """
    Sets up a Jinja2 environment that looks for renderer-specific templates
    first and then generic ones.
    """
    loader = ChoiceLoader([
        FileSystemLoader(renderer_templates),
        FileSystemLoader(Path(__file__).parent / 'templates'),
    ])
    return Environment(
        loader=loader,
        autoescape=select_autoescape()
    )


def copy_static(renderer_static: Path, gallery_path: Path):
    """
    Copies renderer-specific static files first and then generic ones.
    """
    for src in [
        renderer_static,
        Path(__file__).parent / 'static',
    ]:
        shutil.copytree(
            src,
            gallery_path / 'static', dirs_exist_ok=True
        )
    (gallery_path / 'static' / 'README.md').unlink()


def pagination_controls(page_num, total_pages, url_for_page_num):
    controls = []

    def add(page_i, special_page_name=None):
        if special_page_name:
            name = special_page_name
        else:
            name = str(page_i+1)
        enabled = page_i >= 0 and page_i < total_pages and page_i != page_num
        url = url_for_page_num(page_i) if enabled else None
        controls.append({
            "name": name,
            "enabled": enabled,
            "current_page": not (special_page_name) and page_i == page_num,
            "url": url
        })
    # No reason to show any controls at all if there only is one page
    if total_pages == 1:
        return []
    # 0 .. nside page_num nside .. total_pages-1
    n_side = 4
    total_controls = 5 + 2 * n_side
    if total_pages <= total_controls:
        add(page_num - 1, 'Previous')
        for i in range(total_pages):
            add(i)
        add(page_num + 1, 'Next')
        return controls

    div = {
        "name": '..',
        "enabled": False,
        "current_page": False,
        "url": None,
    }

    add(page_num - 1, 'Previous')
    add(0)
    if page_num < 3 + n_side:
        win_begin = 1
        win_end = total_controls - 2
    elif page_num > total_pages - 3 - n_side:
        win_begin = total_pages - total_controls + 2
        win_end = total_pages - 1
    else:
        win_begin = page_num - n_side
        win_end = page_num + n_side + 1

    if win_begin > 2:
        controls.append(div)
    for i in range(win_begin, win_end):
        add(i)
    if win_end < total_pages - 1:
        controls.append(div)
    add(total_pages-1)

    add(page_num + 1, 'Next')
    return controls



import json
import shutil
from pathlib import Path

from jinja2 import (
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    PackageLoader,
    select_autoescape,
)

from gallerator import data_types


def jinja2_env():
    loader = ChoiceLoader(
        [
            PackageLoader("gallerator"),
            FileSystemLoader(Path(__file__).parent / "templates"),
        ]
    )
    return Environment(loader=loader, autoescape=select_autoescape())


class Photoswipe(data_types.Renderer):
    def __init__(self):
        self.jenv = jinja2_env()

    def render(self, template_vars: data_types.TemplateVars):
        template = self.jenv.get_template("page.html")
        vars = template_vars.__dict__.copy()
        vars["data_types"] = data_types
        return template.render(vars)

    def copy_static(self, gallery_path: Path):
        for src in [
            Path(__file__).parent.parent / "static",
            Path(__file__).parent / "static",
        ]:
            shutil.copytree(src, gallery_path / "static", dirs_exist_ok=True)
        (gallery_path / "static" / "README.md").unlink()


def renderer() -> data_types.Renderer:
    return Photoswipe()



#!/usr/bin/env/python3

import sys
from pathlib import Path

"""
Usage:

# If you have sponge from moreutils

cat source.txt | $0 section_name replacement | sponge source.txt

# else

cat source.txt | $0 section_name replacement > source-new.txt
"""

def replace(orig: str, section: str, replacement: str):
    lines = orig.rstrip().split("\n")
    output_lines = []
    inside_section = False
    for line in lines:
        if f'/replace-section:{section}' in line:
            inside_section = False
        elif f'replace-section:{section}' in line:
            inside_section = True
            output_lines.append(line)
            output_lines.append(replacement)
        if not inside_section:
            output_lines.append(line)
    return "\n".join(output_lines)

def main():
    [section, replacement] = sys.argv[1:]
    input = sys.stdin.read()
    print(replace(input, section, replacement))

if __name__ == "__main__":
    main()


from pathlib import Path
import argparse

from . import data_types   
    
class Renderer:
    """Abstract base class for renderers to implement"""

    def render(self, template_vars: data_types.TemplateVars):
        """Returns the contents of the index file for template_vars"""
        raise NotImplementedError

    def copy_static(self, gallery_path: Path):
        """Copies any required files to gallery_path"""
        raise NotImplementedError

    def add_argparse_args(self, parser: argparse.ArgumentParser):
        """The renderer can add any command line arguments it supports. Add them
        in an approriately named group."""
        # Default implementation adds no args
        pass
    
    def update_args(self, args: argparse.Namespace):
        """Stores the command line args for use in other methods. Will be called
        immediately after args have been parsed. """
        self.args = args


from PIL import Image
import math

from . import constants, generated_set


class Thumbnails(generated_set.GeneratedSet):
    """
    Handles generation of thumbnails. All paths in this class, in the API and
    internally, are expected to be hashlib.Path instances
    """

    def __init__(self, generated_dir):
        super().__init__(generated_dir)

    def generated_file_prefix(self):
        return 'thumb'

    def create_file(self, source, destination):
        im = Image.open(source)
        x, y = im.size
        pixels = x * y
        scale = math.sqrt(pixels / constants.thumbnail_target_pixels)
        im.thumbnail((x / scale, y / scale))
        im.save(destination)
